from twisted.trial.unittest import TestCase
from twisted.internet.defer import succeed

import simplejson as json

from txfluiddb.errors import InvalidName
from txfluiddb.client import (
    Namespace, Tag, Endpoint, _HasPath, Object, Blob, BasicCreds)



class PathTests(TestCase):
    """
    Tests for L{_HasPath}.
    """
    def test_brokenPathSubclass(self):
        """
        L{_HasPath.getURL} raises C{NotImplementedError} if C{collectionName}
        is not overridden.
        """
        endpoint = Endpoint('http://fluiddb.test.url/')
        path = _HasPath(u'foo')
        self.assertRaises(NotImplementedError, path.getURL, endpoint)


    def test_nameValidation(self):
        """
        L{InvalidName} is raised if an invalid name is specified.
        """
        self.assertRaises(InvalidName, _HasPath, u'foo bar')
        self.assertRaises(InvalidName, _HasPath, u'foo/bar')
        self.assertRaises(InvalidName, _HasPath, 'foo')
        self.assertRaises(InvalidName, _HasPath, u'foo', u'bar quux')


    def test_getURL(self):
        """
        The getURL method returns the URL for this entity as accessed through
        the given endpoint.
        """
        endpoint = Endpoint('http://fluiddb.test.url/')
        path = _HasPath(u'foo')
        path.collectionName = 'tests'
        self.assertEqual(path.getURL(endpoint),
                         'http://fluiddb.test.url/tests/foo')


    def test_getURLPrefix(self):
        """
        Passing prefix to getURL prepends the given components to the path.
        """
        endpoint = Endpoint('http://fluiddb.test.url/')
        path = _HasPath(u'foo')
        path.collectionName = 'tests'
        self.assertEqual(path.getURL(endpoint, prefix=[u'quux', u'42']),
                         'http://fluiddb.test.url/quux/42/foo')


    def test_getURLSuffix(self):
        """
        Passing suffix to getURL appends the given components to the path.
        """
        endpoint = Endpoint('http://fluiddb.test.url/')
        path = _HasPath(u'foo')
        path.collectionName = 'tests'
        self.assertEqual(path.getURL(endpoint, suffix=[u'quux', u'42']),
                         'http://fluiddb.test.url/tests/foo/quux/42')


    def test_urlEncoding(self):
        """
        URL encoding is done according to IRI rules.
        """
        endpoint = Endpoint('http://fluiddb.test.url/')
        path = _HasPath(u'r\xfasty')
        path.collectionName = 'tests'
        self.assertEqual(path.getURL(endpoint),
                         'http://fluiddb.test.url/tests/r%C3%BAsty')



class NamespaceTests(TestCase):
    """
    Tests for L{Namespace}.
    """
    def setUp(self):
        self.endpoint = MockEndpoint('http://fluiddb.test.url/')
        self.namespace = Namespace(u'test')


    def test_toplevelNamespace(self):
        """
        Constructing a namespace with a single component results in a top-level
        namespace path.
        """
        n = Namespace(u'test')
        self.assertEqual(n.getPath(), u'test')


    def test_childNamespace(self):
        """
        The child method constructs a sub-namespace of the given name.
        """
        n = Namespace(u'test').child(u'child')
        self.assertEqual(n.getPath(), u'test/child')


    def test_pathNamespace(self):
        """
        Constructing a namespace with multiple components results in a
        namespace with the given path.
        """
        n = Namespace(u'test', u'child')
        self.assertEqual(n.getPath(), u'test/child')


    def test_tag(self):
        """
        The tag method returns a tag of the given name in this namespace.
        """
        tag = Namespace(u'test').tag(u'sometag')
        self.assertEqual(tag.getPath(), u'test/sometag')


    def test_collectionName(self):
        """
        Namespaces are under the collection name 'namespaces'.
        """
        namespace = Namespace(u'test')
        self.assertEqual(namespace.child(u'namespace').getURL(self.endpoint),
                         'http://fluiddb.test.url/namespaces/test/namespace')


    def test_getDescription(self):
        """
        Test retrieving a namespace description.
        """
        response = {u'description': u'Testing namespace',
                    u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'}
        self.endpoint.response = json.dumps(response)

        def _gotDesc(description):
            self.assertEqual(description, u'Testing namespace')

        d = self.namespace.getDescription(self.endpoint).addCallback(_gotDesc)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/namespaces/test?returnDescription=true')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_setDescription(self):
        """
        Test setting a namespace description.
        """
        self.endpoint.response = ''

        d = self.namespace.setDescription(self.endpoint, u'New description')
        self.assertEqual(self.endpoint.method, 'PUT')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/namespaces/test')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'description': u'New description'})
        return d


    def test_getNamespaces(self):
        """
        Test retrieving the list of namespaces under a namespace.
        """
        response = {u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'namespaceNames': [u'ns1', u'ns2']}
        self.endpoint.response = json.dumps(response)

        def _gotResult(namespaces):
            self.assertEqual(len(namespaces), 2)
            ns1, ns2 = namespaces
            self.assertIsInstance(ns1, Namespace)
            self.assertEqual(ns1.getPath(), u'test/ns1')
            self.assertIsInstance(ns2, Namespace)
            self.assertEqual(ns2.getPath(), u'test/ns2')

        d = self.namespace.getNamespaces(self.endpoint)
        d.addCallback(_gotResult)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/namespaces/test?returnNamespaces=true')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_getTags(self):
        """
        Test getting the tags in a namespace.
        """
        response = {u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'tagNames': [u'tag1', u'tag2']}
        self.endpoint.response = json.dumps(response)

        def _gotResult(tags):
            self.assertEqual(len(tags), 2)
            tag1, tag2 = tags
            self.assertIsInstance(tag1, Tag)
            self.assertEqual(tag1.getPath(), u'test/tag1')
            self.assertIsInstance(tag2, Tag)
            self.assertEqual(tag2.getPath(), u'test/tag2')

        d = self.namespace.getTags(self.endpoint).addCallback(_gotResult)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/namespaces/test?returnTags=true')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_createChild(self):
        """
        Test creating a new namespace under an existing namespace.
        """
        res = {u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
               u'URI': u'http://fluiddb.test.url/namespaces/test/newchild'}
        self.endpoint.response = json.dumps(res)

        def _gotResult(ns):
            self.assertIsInstance(ns, Namespace)
            self.assertEqual(ns.getPath(), u'test/newchild')

        d = self.namespace.createChild(self.endpoint,
                                       u'newchild',
                                       u'New child namespace')
        self.assertEqual(self.endpoint.method, 'POST')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/namespaces/test')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'name': u'newchild',
                          u'description': u'New child namespace'})
        return d


    def test_createInvalidChild(self):
        """
        Attempting to create a new namespace with an invalid name raises
        L{InvalidName}.
        """
        self.assertRaises(InvalidName,
                          self.namespace.createChild,
                          self.endpoint,
                          u'invalid name',
                          u'some description')


    def test_createTag(self):
        """
        Test creating a tag in a namespace.
        """
        response = {u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'URI': u'http://fluiddb.test.url/tags/test/newtag'}
        self.endpoint.response = json.dumps(response)

        def _gotResult(tag):
            self.assertIsInstance(tag, Tag)
            self.assertEqual(tag.getPath(), u'test/newtag')

        d = self.namespace.createTag(self.endpoint,
                                     u'newtag',
                                     u'New tag',
                                     False)
        self.assertEqual(self.endpoint.method, 'POST')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/tags/test')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'name': u'newtag',
                          u'description': u'New tag',
                          u'indexed': False})
        return d


    def test_createInvalidTag(self):
        """
        Attempting to create a new tag with an invalid name raises
        L{InvalidName}.
        """
        self.assertRaises(InvalidName,
                          self.namespace.createTag,
                          self.endpoint,
                          u'invalid name',
                          u'some description',
                          False)


    def test_delete(self):
        """
        Test deleting a namespace.
        """
        self.endpoint.response = ''

        d = self.namespace.delete(self.endpoint)
        self.assertEqual(self.endpoint.method, 'DELETE')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/namespaces/test')
        return d



class TagTests(TestCase):
    """
    Tests for L{Tag}.
    """
    def setUp(self):
        self.endpoint = MockEndpoint('http://fluiddb.test.url/')
        self.tag = Namespace(u'test').tag(u'sometag')


    def test_collectionName(self):
        """
        Tags are under the collection name 'tags'.
        """
        namespace = Namespace(u'test')
        endpoint = Endpoint('http://fluiddb.test.url/')
        self.assertEqual(namespace.tag(u'tag').getURL(endpoint),
                         'http://fluiddb.test.url/tags/test/tag')


    def test_getDescription(self):
        """
        Test getting a tag description.
        """
        response = {u'description': u'Testing tag',
                    u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'indexed': False}
        self.endpoint.response = json.dumps(response)

        def _gotDesc(description):
            self.assertEqual(description, u'Testing tag')

        d = self.tag.getDescription(self.endpoint).addCallback(_gotDesc)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/tags/test/sometag?returnDescription=true')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_setDescription(self):
        """
        Test setting the tag description.
        """
        self.endpoint.response = ''

        d = self.tag.setDescription(self.endpoint, u'New description')
        self.assertEqual(self.endpoint.method, 'PUT')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/tags/test/sometag')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'description': u'New description'})
        return d


    def test_delete(self):
        """
        Test deleting a tag.
        """
        self.endpoint.response = ''

        d = self.tag.delete(self.endpoint)
        self.assertEqual(self.endpoint.method, 'DELETE')
        self.assertEqual(self.endpoint.url,
                         'http://fluiddb.test.url/tags/test/sometag')
        return d



class MockEndpoint(Endpoint):
    """
    Endpoint with submit mocked out.
    """
    def getPage(self, url, method, postdata=None, headers=None, *a, **kw):
        """
        Store our parameters instead of submitting a request.
        """
        self.url = url
        self.method = method
        self.data = postdata
        self.headers = headers
        rh = self.responseHeaders if hasattr(self, 'responseHeaders') else {}
        rc = self.responseCode if hasattr(self, 'responseCode') else 200
        return succeed((rc, rh, self.response))



class EndpointTests(TestCase):
    """
    Tests for L{Endpoint}.
    """
    def test_anonymousEndpoint(self):
        """
        Constructing an endpoint stores the given base URL.
        """
        url = 'http://fluiddb.test.url/'
        endpoint = Endpoint(url)
        self.assertEqual(endpoint.baseURL, url)


    def test_defaultEndpoint(self):
        """
        Endpoint defaults to the real, global FluidDB endpoint.
        """
        endpoint = Endpoint()
        self.assertTrue(hasattr(endpoint, 'baseURL'))


    def test_getDefaultRoot(self):
        """
        The default root URL has no API version.
        """
        endpoint = Endpoint()
        self.assertEqual(endpoint.getRootURL(), endpoint.baseURL)


    def test_versionedRoot(self):
        """
        The root URL must include the API version if specified.
        """
        endpoint = Endpoint('http://fluiddb.test.url/', '20090817')
        self.assertEqual(endpoint.getRootURL(),
                         'http://fluiddb.test.url/20090817/')


    def test_basicAuth(self):
        """
        An endpoint constructed with basic auth credentials sends them along
        with every request.
        """
        creds = BasicCreds('testuser', 'password')
        endpoint = MockEndpoint('http://fluiddb.test.url/', creds=creds)
        endpoint.response = ''
        endpoint.submit('http://fluiddb.test.url/blah', 'GET')
        self.assertEqual(endpoint.headers['Authorization'],
                         'Basic dGVzdHVzZXI6cGFzc3dvcmQ=')


    def test_unsuppportedAuth(self):
        """
        Constructing an endpoint with an unsupported credentials object raises
        an exception.
        """
        self.assertRaises(ValueError, Endpoint, creds=42)



class ObjectTests(TestCase):
    """
    Tests for L{Object}.
    """
    def setUp(self):
        self.endpoint = MockEndpoint('http://fluiddb.test.url/')


    def test_getTags(self):
        """
        The getTags method returns the about value and the list of tags.
        """
        response = {u'about': u'Testing object',
                    u'tagPaths': [u'test/tag1', u'test/tag2']}
        self.endpoint.response = json.dumps(response)

        def _gotResponse(response):
            about, tags = response
            self.assertEqual(about, u'Testing object')
            self.assertEqual(len(tags), 2)
            tag1, tag2 = tags
            self.assertIsInstance(tag1, Tag)
            self.assertEqual(tag1.getPath(), u'test/tag1')
            self.assertIsInstance(tag2, Tag)
            self.assertEqual(tag2.getPath(), u'test/tag2')

        obj = Object(u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        d = obj.getTags(self.endpoint).addCallback(_gotResponse)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'showAbout': True})
        return d


    def makeGetRequest(self, response):
        """
        Utility wrapper for testing L{Object.get} operations.
        """
        value, contentType = response
        self.endpoint.response = value
        self.endpoint.responseHeaders = { 'content-type' : contentType, }
        obj = Object(u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        tag = Namespace(u'test').child(u'tag')
        d = obj.get(self.endpoint, tag)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(self.endpoint.data, None)
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx/test/tag')
        return d


    def test_getString(self):
        """
        Retrieving a string value results in a C{unicode} object.
        """
        response = (json.dumps('Data goes where?'),
                    'application/vnd.fluiddb.value+json')
        def _gotResponse(response):
            self.assertIsInstance(response, unicode)
            self.assertEqual(response, u'Data goes where?')
        return self.makeGetRequest(response).addCallback(_gotResponse)


    def test_getInteger(self):
        """
        Retrieving a integer value results in an integer object.
        """
        response = (json.dumps(42), 'application/vnd.fluiddb.value+json')
        def _gotResponse(response):
            self.assertEqual(response, 42)
        return self.makeGetRequest(response).addCallback(_gotResponse)


    def test_getBlob(self):
        """
        Retrieving a binary value results in a C{Blob} object.
        """
        response = ('<p>foo</p>', 'text/html')
        def _gotResponse(response):
            self.assertEqual(response.contentType, u'text/html')
            self.assertEqual(response.data, '<p>foo</p>')
        return self.makeGetRequest(response).addCallback(_gotResponse)


    def test_deleteTag(self):
        """
        Deleting a tag returns nothing if successful.
        """
        self.endpoint.response = ''

        obj = Object(u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        tag = Namespace(u'test').child(u'tag')
        d = obj.delete(self.endpoint, tag)
        self.assertEqual(self.endpoint.method, 'DELETE')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx/test/tag')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_setTag(self):
        """
        Test setting the value of a tag on an object.
        """
        self.endpoint.response = ''

        obj = Object(u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        tag = Namespace(u'test').child(u'tag')
        d = obj.set(self.endpoint, tag, 42)
        self.assertEqual(self.endpoint.method, 'PUT')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx/test/tag')
        self.assertEqual(json.loads(self.endpoint.data), 42)
        return d


    def test_setBlob(self):
        """
        Test setting the value of a tag on an object to a blob.
        """
        self.endpoint.response = ''

        obj = Object(u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')
        tag = Namespace(u'test').child(u'tag')
        blob = Blob(u'text/html', '<p>foo</p>')
        d = obj.setBlob(self.endpoint, tag, blob)
        self.assertEqual(self.endpoint.method, 'PUT')
        self.assertEqual(
            self.endpoint.url,
            'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx/test/tag')
        self.assertEqual(self.endpoint.data, '<p>foo</p>')
        self.assertEqual(self.endpoint.headers['Content-Type'], 'text/html')
        return d


    def test_createObject(self):
        """
        Creating an object returns the newly created object.
        """
        response = {u'URI': u'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'}
        self.endpoint.response = json.dumps(response)

        def _gotResponse(obj):
            self.assertIsInstance(obj, Object)
            self.assertEqual(obj.uuid, u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')

        d = Object.create(self.endpoint).addCallback(_gotResponse)
        self.assertEqual(self.endpoint.method, 'POST')
        self.assertEqual(self.endpoint.url, 'http://fluiddb.test.url/objects')
        self.assertEqual(self.endpoint.data, None)
        return d


    def test_createObjectWithAbout(self):
        """
        Test creating an object with an about value.
        """
        response = {u'URI': u'http://fluiddb.test.url/objects/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx',
                    u'id': u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'}
        self.endpoint.response = json.dumps(response)

        def _gotResponse(obj):
            self.assertIsInstance(obj, Object)
            self.assertEqual(obj.uuid, u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx')

        d = Object.create(self.endpoint, u'New object').addCallback(_gotResponse)
        self.assertEqual(self.endpoint.method, 'POST')
        self.assertEqual(self.endpoint.url, 'http://fluiddb.test.url/objects')
        self.assertEqual(json.loads(self.endpoint.data),
                         {u'about': u'New object'})
        return d


    def test_query(self):
        """
        Querying returns all objects matching the query.
        """
        response = {u'ids': [u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxx1',
                             u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxx2']}
        self.endpoint.response = json.dumps(response)

        def _gotResponse(objs):
            self.assertEqual(len(objs), 2)
            obj1, obj2 = objs
            self.assertIsInstance(obj1, Object)
            self.assertEqual(obj1.uuid, u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxx1')
            self.assertIsInstance(obj2, Object)
            self.assertEqual(obj2.uuid, u'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxx2')

        d = Object.query(self.endpoint, u'has fluiddb/about').addCallback(_gotResponse)
        self.assertEqual(self.endpoint.method, 'GET')
        self.assertEqual(self.endpoint.url, 'http://fluiddb.test.url/objects?query=has+fluiddb%2Fabout')
        self.assertEqual(self.endpoint.data, None)
        return d
